/*
 * Copyright (c) 2009, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * The names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <QMainWindow>
#include <QMessageBox>
#include <QProgressBar>
#include <QProgressDialog>
#include <QLabel>
#include <QDir>
#include <QCloseEvent>
#include <QSettings>
#include <QFont>

#include "main_widget.h"
#include "puzzletrs_widget.h"
#include "settings_widget.h"
#include "../logic/defines.h"

MainWidget::MainWidget(QWidget *parent)
	: QMainWindow(parent)
{
	setupUi(this);
	spth = new SpellathonCls(this);
	settings = new QSettings("Ephemera", "Spellathon");

	puzzleLetrs_ = new PuzzleLetrs(puzzleBox);
	statWordCnt = new QLabel(this);
	statMsg = new QLabel(this);
	statWordCnt->setText("          ");
	statusOutput->insertWidget(0, statMsg, 1);
	statusOutput->insertPermanentWidget(1, statWordCnt, 0);
	spbar = new QProgressBar(statusOutput);
	spbar->setTextVisible(false);
	spbar->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Ignored);
	statusOutput->insertWidget(1, spbar, 0);
	spbar->hide();
	playButton->setEnabled(false);
	solveButton->setEnabled(false);
	hintButton->setEnabled(false);
	finishButton->setEnabled(false);
	wordInput->setEnabled(false);

	connect(action_Exit, SIGNAL(triggered()), 
			this, SLOT(close()));
	connect(action_Instructions, SIGNAL(triggered()), 
			this, SLOT(showInst()));
	connect(action_About, SIGNAL(triggered()), 
			this, SLOT(showAbout()));
	connect(action_Settings, SIGNAL(triggered()),
			this, SLOT(openSettings()));

	connect(exitButton, SIGNAL(clicked()), 
			this, SLOT(close()));
	connect(playButton, SIGNAL(clicked()), 
			this, SLOT(playPressed()));
	connect(hintButton, SIGNAL(clicked()), 
			wordInput, SLOT(setFocus()));
	connect(hintButton, SIGNAL(clicked()), 
			spth, SLOT(hint()));
	connect(finishButton, SIGNAL(clicked()), 
			this, SLOT(finishPressed()));
	connect(solveButton, SIGNAL(clicked()), 
			this, SLOT(solvePressed()));
	connect(wordInput, SIGNAL(returnPressed()), 
			this, SLOT(wordEntered()));
	connect(wordsOutput, SIGNAL(selectionChanged()),
			wordsOutput, SLOT(copy()));

	connect(spth, SIGNAL(printWords(QString)), 
			wordsOutput, SLOT(setText(QString)));
	connect(spth, SIGNAL(printWordsFmt(QString)), 
			wordsOutput, SLOT(setHtml(QString)));
	connect(spth, SIGNAL(printStatus(QString)), 
			statMsg, SLOT(setText(QString)));
	connect(spth, SIGNAL(printStatusTemp(QString, int)), 
			statusOutput, SLOT(showMessage(QString, int)));
	connect(spth, SIGNAL(printMsg(QString, QString)), 
			this, SLOT(printMsg(QString, QString)));
	connect(spth, SIGNAL(printWordCnt(QString)),
			statWordCnt, SLOT(setText(QString)));
	connect(spth, SIGNAL(fatalMsg(QString)), 
			this, SLOT(fatalMsg(QString)));
	connect(spth, SIGNAL(disableHints()), 
			this, SLOT(disableHints()));
	connect(spth, SIGNAL(printPuzzle(QString)), 
			this, SLOT(printPuzzle(QString)));
	connect(spth, SIGNAL(finished()), 
			this, SLOT(finishPressed()));
	connect(spth, SIGNAL(noDict(QString)), 
			this, SLOT(noDict(QString)));
	connect(spth, SIGNAL(spellathonReady()),
			this, SLOT(spellathonReady()));
	connect(spth, SIGNAL(wordchkimg(bool)),
			wordInput, SLOT(chk(bool)));

	connect(this, SIGNAL(play()), 
			spth, SLOT(play()));
	connect(this, SIGNAL(finish()), 
			spth, SLOT(finish()));
	connect(this, SIGNAL(setWord(QString)), 
			spth, SLOT(setWord(QString)));
	connect(this, SIGNAL(solve(QString)), 
			spth, SLOT(solve(QString)));
	connect(this, SIGNAL(setSolRange(int, int)), 
			spth, SLOT(setSolRange(int, int)));
	connect(this, SIGNAL(setDoScore(bool)),
			spth, SLOT(setDoScore(bool)));
	connect(this, SIGNAL(setDoTime(bool)),
			spth, SLOT(setDoTime(bool)));
	connect(this, SIGNAL(wpl_update(int)), 
			spth, SLOT(wpl_update(int)));

	pendingDictChange = "";

	resize(settings->value("mainwidget/size", size()).toSize());
	move(settings->value("mainwidget/pos", pos()).toPoint());

	QFont monoFont = settings->value("font/mono", wordsOutput->font()).value<QFont>();
	wordsOutput->setFont(monoFont);
	puzzleLetrs_->setFont(monoFont);
}

void
MainWidget::init()
{	
	puzzleLetrs_->move((puzzleBox->size().width() - puzzleLetrs_->size().width())/2,
					   (puzzleBox->size().height() - puzzleLetrs_->size().height())/2);
	puzzleLetrs_->show();

	if (!settings->contains("dict/precomp"))
		settings->setValue("dict/precomp", PRECOMP);

	if (!settings->contains("mainwidget/size"))
		settings->setValue("mainwidget/size", size());
	if (!settings->contains("mainwidget/pos"))
		settings->setValue("mainwidget/pos", pos());
	QDir dp;
	if (!settings->contains("dict/path"))
		settings->setValue("dict/path", dp.absoluteFilePath(DICTPATH));
	if (!settings->contains("font/mono"))
		settings->setValue("font/mono", wordsOutput->font());
	if (!settings->contains("score/show"))
		settings->setValue("score/show", true);
	if (!settings->contains("time/show"))
		settings->setValue("time/show", true);
	if (!settings->contains("solutions/min"))
		settings->setValue("solutions/min", MIN_SOL);
	if (!settings->contains("solutions/max"))
		settings->setValue("solutions/max", MAX_SOL);
	settings->sync();

	if (settings->status() != QSettings::NoError)
		fatalMsg(QString("Error updating settings: %1").arg(settings->fileName()));

	settingsW_ = new SettingsW(this, settings);
	if (settings->status() != QSettings::NoError)
		fatalMsg(QString("Error updating settings: %1").arg(settings->fileName()));

	int sz = wordsOutput->size().width();
	int wsz = wordsOutput->fontMetrics().width("zzzzzzzzz");
	spth->init(settings->value("dict/path").toString(), 
			   settings->value("dict/precomp").toBool(), (sz-wsz/3)/wsz);

	emit setSolRange(settings->value("solutions/min").toInt(), 
					 settings->value("solutions/max").toInt());
	emit setDoScore(settings->value("score/show").toBool());
	emit setDoTime(settings->value("time/show").toBool());
}

void
MainWidget::openSettings()
{
	int r = settingsW_->exec();
	if (settings->status() != QSettings::NoError)
		fatalMsg(QString("Error updating settings: %1").arg(settings->fileName()));
	if (r) {
		emit setDoScore(settings->value("score/show").toBool());
		emit setDoTime(settings->value("time/show").toBool());
		emit setSolRange(settings->value("solutions/min").toInt(),
						 settings->value("solutions/max").toInt());

		QString d = settings->value("dict/path").toString();
		bool p = settings->value("dict/precomp").toBool();
		if (d != spth->dict() || p != spth->isPrecomp()) {
			if (spth->isIngame())
				pendingDictChange = d;
			else
				spth->init(d, p);
		}
		QFont monoFont = settings->value("font/mono").value<QFont>();
		if (monoFont != wordsOutput->font()) {
			wordsOutput->setFont(monoFont); 
			puzzleLetrs_->setFont(monoFont);
			int sz = wordsOutput->size().width();
			int wsz = wordsOutput->fontMetrics().width("zzzzzzzzz");
			emit wpl_update((sz-wsz/3)/wsz);
		}
	}
}

void 
MainWidget::showAbout()
{
	const char* svn_version();
	QMessageBox::about(this, "About", QString("Spellathon "VERSION" (svn: %1)"
				 "\n\nCopyright (c) 2009 Girish Rao" 
				 "\n\nwww.sourceforge.net/projects/spellathon").arg(svn_version()));
}

void 
MainWidget::showInst()
{
	QDialog *d = new QDialog(this);
	QTextEdit *t = new QTextEdit(d);
	QPushButton *p = new QPushButton(d);
	QVBoxLayout *l = new QVBoxLayout(d);

	connect(p, SIGNAL(clicked()), d, SLOT(close()));
	d->move(this->pos().x() + this->frameSize().width(), this->pos().y());
	d->resize(380, 500);
	d->setWindowTitle("Instructions");
	t->setReadOnly(true);
	l->addWidget(t);
	l->addWidget(p, 0, Qt::AlignCenter);
	p->setText("&Ok");
	t->setLineWrapMode(QTextEdit::WidgetWidth);
	t->setHtml(
		QString("<br><b>Game instructions</b><br><br>"
		   "-&nbsp;Find all the words of four or more letters"
		   " from the letters shown in the puzzle.<br><br>"
		   "-&nbsp;In making a word, each letter may be used "
		   "once only.<br><br>-&nbsp;Each word must contain the "
		   "central letter.<br><br>-&nbsp;There should be atleast"
		   " one seven letter word.<br><br>-&nbsp;Plurals, foreign words and"
		   " proper names are not allowed.<br><br>"
		   "<br><b>Notes</b><br><br>"
		   "Hints will decrease your score.<br><br>"
		   "To get the top score/rating you will "
		   "also need to be fast.<br><br>"
		   "Ratings: %1, %2, %3, %4, %5, %6.<br><br>"
		   "To solve a custom puzzle enter the letters of puzzle "
		   "in the corresponding boxes and press Solve.<br><br>"
		   "Use the Settings->Precompute option if you are using "
		   "an unusual solution range and noticing delays (on slow computers)."
		   "<br>This will calculate and store some game information on the disk"
		   " so that it doesn't need to be computed at runtime."
		   ).arg(Ratings[0]).arg(Ratings[1]).arg(Ratings[2])
			.arg(Ratings[3]).arg(Ratings[4]).arg(Ratings[5])
		);
	d->show();
}
	
void 
MainWidget::disableHints()
{
	hintButton->setEnabled(false);
	statusOutput->showMessage("No more hints.", 3000);
}

void
MainWidget::noDict(QString err)
{
	playButton->setEnabled(false);
	hintButton->setEnabled(false);
	solveButton->setEnabled(false);
	finishButton->setEnabled(false);
	puzzleLetrs_->clear();
	puzzleLetrs_->setReadOnly(false);
	wordInput->clear();
	wordsOutput->clear();
	puzzleLetrs_->centralLetr()->setFocus();
	QMessageBox::critical(this, "Dictionary", err);
}

void
MainWidget::spellathonReady()
{
	playButton->setEnabled(true);
	solveButton->setEnabled(true);
	playButton->setFocus();
}

void 
MainWidget::playPressed()
{
	if (spth->isInprogress()) {
		int r = QMessageBox::warning(this, "New game", 
					     	"Game in progress. Quit current game?", 
					     	QMessageBox::Cancel | 
					     	QMessageBox::Ok, 
						QMessageBox::Cancel);

		if (r == QMessageBox::Cancel)
			return;
	}

	if (pendingDictChange != "") {
		spth->init(pendingDictChange, settings->value("dict/precomp").toBool());
		pendingDictChange = "";
		if (!spth->isDictOpen())
			return;
	}

	puzzleLetrs_->clear();

	wordsOutput->setText("");
	wordInput->setEnabled(false);
	wordInput->clear();
	statMsg->clear();
	statWordCnt->setText("          ");
	playButton->setEnabled(false);
	hintButton->setEnabled(false);
	finishButton->setEnabled(false);
	solveButton->setEnabled(false);
	action_Settings->setEnabled(false);
	
	emit play();

	action_Settings->setEnabled(true);
	playButton->setEnabled(true);

	if (spth->isIngame()) {
		hintButton->setEnabled(true);
		finishButton->setEnabled(true);
		wordInput->setEnabled(true);
		wordInput->setFocus();
		wordInput->setInitState(true);
		puzzleLetrs_->setReadOnly(true);
	} else {
		puzzleLetrs_->setReadOnly(false);
		solveButton->setEnabled(true);
	}
}

void 
MainWidget::fatalMsg(QString s)
{
	QMessageBox::critical(this, "Error", s);
}
	
void 
MainWidget::printMsg(QString t, QString s)
{
	QMessageBox::information(this, t, s);
}

void 
MainWidget::finishPressed()
{
	playButton->setEnabled(true);
	hintButton->setEnabled(false);
	finishButton->setEnabled(false);
	solveButton->setEnabled(true);
	playButton->setFocus();
	wordInput->clear();
	wordInput->setEnabled(false);

	puzzleLetrs_->setReadOnly(false);

	emit finish();
}

void
MainWidget::solvePressed()
{
	wordsOutput->setText("");
	wordInput->setEnabled(false);
	statWordCnt->setText("          ");
	statMsg->clear();

	if (pendingDictChange != "") {
		spth->init(pendingDictChange, settings->value("dict/precomp").toBool());
		pendingDictChange = "";
	}

	QString p = puzzleLetrs_->letrs();
	if (p != "") {
		emit solve(p);
		playButton->setFocus();
	} else {
		puzzleLetrs_->clear();
		QMessageBox::information(this, "Solve puzzle", 
					 "Please enter the seven letters of the puzzle in the"
					 " corresponding boxes and then press Solve.");
		puzzleLetrs_->centralLetr()->setFocus();
	}
	return;
}

void 
MainWidget::wordEntered()
{
	emit setWord(wordInput->text().toLower());
	wordInput->clear();
}

void 
MainWidget::printPuzzle(QString s)
{
	puzzleLetrs_->setLetrs(s);
	puzzleLetrs_->setReadOnly(true);
}

void
MainWidget::resizeEvent(QResizeEvent *event)
{
	int sz = event->size().width();
	int wsz = wordsOutput->fontMetrics().width("zzzzzzzzz");
	emit wpl_update((sz-wsz/3)/wsz);
	puzzleLetrs_->move((puzzleBox->size().width() - puzzleLetrs_->size().width())/2,
					   (puzzleBox->size().height() - puzzleLetrs_->size().height())/2);
	event->accept();
}

void
MainWidget::closeEvent(QCloseEvent *event)
{
	settings->setValue("mainwidget/size", size());
	settings->setValue("mainwidget/pos", pos());
	event->accept();
}

MainWidget::~MainWidget()
{
	delete spth;
	delete settings; 
}

void
MainWidget::statProgressOpen(int min, int max)
{
	spbar->setRange(min, max);
	spbar->show();
}

void
MainWidget::statProgressClose()
{
	spbar->hide();
}

void
MainWidget::statProgressUpdate(int n)
{
	spbar->setValue(n);
}

void
MainWidget::progressDiagOpen(const QString& txt, const QString& ttl, int min, int max)
{
	pbar = new QProgressDialog(txt, 0, min, max, this);
	pbar->setWindowModality(Qt::WindowModal);
	pbar->setMinimumDuration(0);
	pbar->setWindowTitle(ttl);
}

void
MainWidget::progressDiagUpdate(int n)
{
	pbar->setValue(n);
}

void
MainWidget::progressDiagClose()
{
	delete pbar;
	pbar = 0;
}

