/*
 * Copyright (c) 2009, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * The names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <new>
#include <cstdlib>
#include <cassert>
#include <iostream>
#include <errno.h>
#include <time.h>

#include <QTimer>

#include "spellathon_cls.h"
#include "main_widget.h"
#include "../logic/defines.h"
#include "../logic/dict.h"
#include "../logic/util.h"
#include "../logic/game.h"
#include "../logic/dispwordsbuf.h"
#include "../logic/score.h"

namespace Spellathon
{
	
	SpellathonCls::SpellathonCls(MainWidget *g)
		:	gui_h_(g), dct(0), dsp(0), game(0),
			minSol_(MIN_SOL), maxSol_(MAX_SOL),
			ingame(false), inprogress(false),
			dictpath(""), dict_opened(false),
			precomp(false), t(0)
	{ 
			timer = new QTimer(this);
			timer->setInterval(1000);
			connect(timer, SIGNAL(timeout()), this, SLOT(updateTime()));
	}

	void
	SpellathonCls::setSolRange(int min, int max)
	{
		maxSol_ = max;
		minSol_ = min;
	}

	void 
	SpellathonCls::setDoScore(bool e)
	{
		doscore = e;
	}

	void 
	SpellathonCls::setDoTime(bool e)
	{
		dotime = e;
	}

	void 
	SpellathonCls::wpl_update(int n)
	{
		if (dsp && n != wpl_) {
			dsp->setWpl(wpl_ = n);
			emit printWords(dsp->bufHtml());
		}
	}

	void
	SpellathonCls::init(QString dpath, bool pc, int wpl)
	{
		destroy();

		dictpath = dpath;
		dict_opened = false;
		ingame = false;
		inprogress = false;
		precomp = pc;
		wpl_ = wpl;
		try {
			dct = new Dict();
			connect(dct, SIGNAL(progressbarOpen(const QString&, 
					const QString&, int, int)), 
					gui_h_, SLOT(progressDiagOpen(const QString&, 
					const QString&, int, int)));
			connect(dct, SIGNAL(progressbarUpdate(int)), 
					gui_h_, SLOT(progressDiagUpdate(int)));
			connect(dct, SIGNAL(progressbarClose()), 
					gui_h_, SLOT(progressDiagClose()));

			try {
				dct->init(dictpath.toAscii().data(), precomp);
			} 
			catch (string s) 
			{
				emit noDict(s.c_str());
				destroy();
				return;
			}
	
			try {	
				game = new Game(dct);
				dsp = new DispWordsBuf;
			} 
			catch (string s) 
			{
				emit fatalMsg(s.c_str());
				exit(1);
			}
			connect(game, SIGNAL(progressbarOpen(int, int)), 
					gui_h_, SLOT(statProgressOpen(int, int)));
			connect(game, SIGNAL(progressbarClose()), 
					gui_h_, SLOT(statProgressClose()));
			connect(game, SIGNAL(progressbarUpdate(int)), 
					gui_h_, SLOT(statProgressUpdate(int)));
		}
		catch (std::bad_alloc &ba)
		{
			string s = "Can't allocate memory: ";
		    s += ba.what();
			emit fatalMsg(s.c_str());
			exit(1);
		}

		dict_opened = true;
		emit spellathonReady();	
	}
	
	void
	SpellathonCls::destroy()
	{
		delete dsp;		dsp = 0;
		delete game;	game = 0;
		delete dct;		dct = 0;
	}
	
	void
	SpellathonCls::play()
	{
		assert(dictpath != "");

		inprogress = false;
		
		timer->stop();
		bool r = game->play(precomp, minSol_, maxSol_);
		if (r) {
			ingame = true;
			emit printPuzzle(game->sig().c_str());
			dsp->init(game->words(), wpl_);
			emit printWords(dsp->bufHtml());
			emit printWordCnt(QString("Words:  0 / %1 ").arg(dsp->nWords()));
			t = 0;
			timer->start();
			if (dotime)
				emit printStatus("Time:  0m 0s");
		} else {
			ingame = false;
			printMsg("Game", "No suitable game found.\n\n"
			       	 "Choose a different Solution range or use "
					 "a different dictionary from: File->Settings.");
		}
	}
	
	void
	SpellathonCls::updateTime()
	{
		++t;
		if (dotime)
			emit printStatus(QString("Time:  %1m %2s")
					 .arg((int)t/60).arg((int)t%60));
		else
			printStatus("");
	}
		
	void
	SpellathonCls::hint()
	{
		assert(ingame);
	
		dsp->setHint();
		emit printWords(dsp->bufHtml());
	
		if (dsp->hintLevel() >= 2)
			emit disableHints();
	}
	
	void
	SpellathonCls::finish()
	{
		assert(ingame);

		timer->stop();

		QString status("");

		if (dotime) 
			status.append(QString("Time:  %1m %2s      ").arg((int)t/60).arg((int)t%60));

		if (doscore) {
			Score sc(dsp->nSetWords(), dsp->nWords(), 
				 dsp->hintLevel(), t);

			status.append(QString("Score:  %3").arg(sc.score()));

			emit printMsg("Score", 
				      QString("Correct words:  %1 / %2           Time: %5m %6s"
					      "\n\nScore:  %3    [%4]        \n")
				      .arg(dsp->nSetWords()).arg(dsp->nWords())
				      .arg(sc.score()).arg(sc.rating().c_str()).arg((int)t/60).arg((int)t%60));
		}

		emit printStatus(status);

		dsp->setSol();
		emit printWordsFmt(dsp->bufHtml());

		ingame = false;
		inprogress = false;
	}
	
	void
	SpellathonCls::solve(QString str)
	{
		assert(dictpath != "" && !ingame);

		int n = game->solve(str.toStdString());
		if (n > 0) {
			dsp->init(game->words(), wpl_);
			dsp->setSol();
			emit printWords(dsp->bufHtml());
			emit printWordCnt(QString("Words:  %1 ").arg(dsp->nWords()));
		} else {
			emit printWordCnt("Words:  0 ");
			emit printStatus("No Solution for given puzzle.");
		}
	}
	
	void
	SpellathonCls::setWord(QString w)
	{
		assert(ingame);
	
		switch (dsp->setWord(w.toStdString())) {
		case 1:
			inprogress = true;

			emit printWordCnt(QString("Words:  %1 / %2 ")
					  .arg(dsp->nSetWords())
					  .arg(dsp->nWords()));

			emit printWords(dsp->bufHtml());
			emit wordchkimg(true);

			if (dsp->nWords() == dsp->nSetWords()) 
				emit finished();
			
			break;

		case -1:
			emit wordchkimg(false);
			break;

		case 0:
			emit printStatusTemp("Word already entered.", 3000);
			break;
		}
			
	}

}

