/*
 * Copyright (c) 2009, Girish Rao
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * The names of its contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT i
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#include <cassert>

#include "solver.h"
#include "util.h"
#include "dict.h"

namespace Spellathon
{
	
	/* performance critical routine */
	int
	Solver::solve(const char* letrs, int min, int max)
	{
		char	signarr[NCOMBINATIONS][MAX_LETRS+1];
	
		int c = signGen(letrs, signarr);
		bool has7 = false;
	
		wordPtrs.clear();
		vector<char *> w;
		w.reserve(64);
		/* 
		 * for all the sign. generated by signgen() find matching 
		 * sign (and thus words) from the dictionary
		 */
		for (int k = 0; k < c; k++) {
			w.clear();
			dict->matchSign(signarr[k], w);
										
			/* remove plurals */
			int i = (int)w.size();
			while (--i >= 0) {
				int slen = strlen(w[i]);
				int slen2 = slen - 1;
				if (w[i][slen2] == 's') {
					/* note: promiscuous pointer manipulation below */
					w[i][slen2] = 0;	
					if (dict->matchWord(w[i])) {
						w[i][slen2] = 's';
						continue;
					} else {
						w[i][slen2] = 's';
					}
				}
				if (slen == 7)
					has7 = true;

				wordPtrs.push_back(w[i]); 
			}

			if ((int)wordPtrs.size() > max)
				return -1;
		}
	
		if ((int)wordPtrs.size() < min)
			return -1;
	
		if (!has7) 
			return -1;

		return wordPtrs.size();
	}
	
	/* 
	 * find all the letter combinations of strlen 4, 5, 6 and 7.
	 * the letters can be duplicates so that needs to be handled.
	 * since this a fixed problem i have used a table lookup.
	 */
	/* performance critical routine */
	int
	Solver::signGen(const char *letters, char signarr[][MAX_LETRS+1])
	{
		int		i, j, k;
		const uint8_t	smask[NCOMBINATIONS] = { 
			/* C(6,3) */
			0x7,  0xb,  0xd,  0xe,  0x13, 0x15, 
			0x16, 0x19, 0x1a, 0x1c, 0x23, 0x25, 
			0x26, 0x29, 0x2a, 0x2c, 0x31, 0x32, 
			0x34, 0x38,
			/* C(6,4) */
			0xf,  0x17, 0x1b, 0x1d, 0x1e, 0x27, 
			0x2b, 0x2d, 0x2e, 0x33, 0x35, 0x36, 
			0x39, 0x3a, 0x3c,
			/* C(6,5) */
			0x1f, 0x2f, 0x37, 0x3b, 0x3d, 0x3e,
			/* C(6,6) */
			0x3f 
		}; 
		int8_t coff[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
	
		memset(signarr, 0, NCOMBINATIONS * (MAX_LETRS+1));

		int n = 0;

		char letrlookup[64];
		letrlookup[0] = 0;
		letrlookup[0x1] = letters[0];
		letrlookup[0x2] = letters[1];
		letrlookup[0x4] = letters[2];
		letrlookup[0x8] = letters[3];
		letrlookup[0x10] = letters[4];
		letrlookup[0x20] = letters[5];
		/* 
		 * the central letter is present in all the words so
		 * find the combinations of the remaining 6 letters 
		 */
		for (i = 0; i < NCOMBINATIONS; i++) {
			/* generate a letter combination from the bitmask */
			j = 0;
			signarr[n][j] = letrlookup[smask[i]&1]; 	 j += smask[i] & 1;
			signarr[n][j] = letrlookup[smask[i]&(1<<1)]; j += (smask[i] & (1<<1))>>1;
			signarr[n][j] = letrlookup[smask[i]&(1<<2)]; j += (smask[i] & (1<<2))>>2;
			signarr[n][j] = letrlookup[smask[i]&(1<<3)]; j += (smask[i] & (1<<3))>>3;
			signarr[n][j] = letrlookup[smask[i]&(1<<4)]; j += (smask[i] & (1<<4))>>4;
			signarr[n][j] = letrlookup[smask[i]&(1<<5)]; j += (smask[i] & (1<<5))>>5;
	
			/* add the central letter */
			signarr[n][j++] = letters[MAX_LETRS-1];
	
			/* sign it */
			sign(signarr[n], j);
	
			if (coff[j] == -1)
			   coff[j] = n;

			/* ignore duplicate comb. in the case where letters 
			 * are duplicates */
			k = n-1;
			while (k >= coff[j]) {
				if (strcmp_em(signarr[n], signarr[k])) {
					n--;	
					break;
				}
				k--;
			}
			n++;
		}
	
		return n;	
	}

	const vector<string> & 
	Solver::words() 
	{ 
		words_.clear();

		for (int i = 0; i < (int)wordPtrs.size(); i++) 
			words_.push_back(wordPtrs[i]); 

		return words_;
	}
}
